# Generated by Django 3.1.5 on 2021-09-29 12:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('dkron', '0001_initial_20210729'),
    ]

    operations = [
        migrations.AddField(
            model_name='job',
            name='retries',
            field=models.IntegerField(default=0),
        ),
    ]



import re
from urllib.parse import parse_qsl

from django import forms
from django.contrib import admin
from django.http.response import HttpResponseForbidden, HttpResponseRedirect
from django.templatetags.static import static
from django.urls import reverse
from django.utils.html import format_html
from dkron import models, utils
from dkron.forms import JobForm


class JobAdminForm(forms.ModelForm):
    NAME_RE = re.compile(r'^[a-zA-Z0-9\-_]+$')

    def clean_name(self):
        if self.NAME_RE.match(self.cleaned_data['name']) is None:
            raise forms.ValidationError('Invalid value')
        return self.cleaned_data['name'].lower()

    def clean_schedule(self):
        # crappy way to validate it but oh well...
        j = models.Job(
            name='tmp_test_job', schedule=self.cleaned_data['schedule'], enabled=False, command='echo validator'
        )
        try:
            utils.sync_job(j)
        except utils.DkronException:
            raise forms.ValidationError('Invalid value')
        else:
            try:
                utils.delete_job(j)
            except Exception:
                # don't really care, just cleanup
                pass
        return self.cleaned_data['schedule']


@admin.register(models.Job)
class JobAdmin(admin.ModelAdmin):
    form = JobAdminForm
    list_display = (
        'name',
        'schedule',
        'command',
        'description',
        'enabled',
        'notify_on_error',
        'get_last_run',
        'get_dkron_link',
    )
    list_display_links = ('name',)
    search_fields = ('name', 'schedule', 'command', 'description')
    list_filter = ('name', 'schedule', 'enabled', 'last_run_success', 'notify_on_error')
    actions = ['disable_jobs', 'enable_jobs']
    form = JobForm

    def has_dashboard_permission(self, request):
        return request.user.has_perm('dkron.can_use_dashboard')

    def _change_job_enabled(self, request, queryset, value):
        queryset.update(enabled=value)
        for job in queryset:
            try:
                utils.sync_job(job)
            except utils.DkronException as e:
                self.message_user(request, f'Failed to sync {job.name} config with dkron - {str(e)}', 'ERROR')

    def get_dkron_link(self, obj):
        return format_html(
            '''
            <a href="{}" target="_blank" rel="noopener">
            <i class="fas fa-external-link-alt"></i>
            </a>
            ''',
            utils.job_executions(obj.name),
        )

    get_dkron_link.short_description = 'Dkron Link'

    def get_last_run(self, obj):
        if obj.last_run_date is None:
            return None
        return format_html(
            '{} <img src="{}">',
            obj.last_run_date.strftime('%d/%m/%Y %H:%M'),
            static('admin/img/icon-{}.svg'.format('yes' if obj.last_run_success else 'no')),
        )

    get_last_run.short_description = 'Last Run'
    get_last_run.admin_order_field = 'last_run_date'

    def disable_jobs(self, request, queryset):
        self._change_job_enabled(request, queryset, False)

    disable_jobs.short_description = 'Disable selected jobs'

    def enable_jobs(self, request, queryset):
        self._change_job_enabled(request, queryset, True)

    enable_jobs.short_description = 'Enable selected jobs'

    def save_model(self, request, obj, form, change):
        super().save_model(request, obj, form, change)
        try:
            utils.sync_job(obj, job_update=None)
        except utils.DkronException as e:
            self.message_user(request, f'Failed to sync {obj.name} config with dkron - {str(e)}', 'ERROR')

    def delete_model(self, request, obj):
        super().delete_model(request, obj)
        try:
            utils.delete_job(obj)
        except utils.DkronException as e:
            self.message_user(request, f'Failed to delete {obj.name} from dkron - {str(e)}', 'ERROR')

    def resync(self, request):
        if not self.has_dashboard_permission(request):
            return HttpResponseForbidden()
        c = [0, 0]
        for _, action, result in utils.resync_jobs():
            if result:
                self.message_user(request, result, 'ERROR')
            else:
                if action == 'u':
                    c[0] += 1
                else:
                    c[1] += 1

        self.message_user(request, '%d jobs updated and %d deleted' % tuple(c))
        post_url = reverse('admin:dkron_job_changelist', current_app=self.admin_site.name)
        preserved_filters = self.get_preserved_filters(request)
        preserved_filters = dict(parse_qsl(preserved_filters)).get('_changelist_filters')
        if preserved_filters:
            post_url = '%s?%s' % (post_url, preserved_filters)
        return HttpResponseRedirect(post_url)

    def get_urls(self):
        from django.urls import path

        urls = super().get_urls()
        urls.insert(0, path('resync/', self.admin_site.admin_view(self.resync), name='dkron_job_resync'))
        return urls

    def changelist_view(self, request, extra_context=None):
        extra_context = extra_context or {}
        extra_context['has_dashboard_permission'] = self.has_dashboard_permission(request)
        extra_context['has_change_permission'] = self.has_change_permission(request)
        return super().changelist_view(request, extra_context)



from django.shortcuts import reverse
from django.apps import AppConfig
from django.conf import settings

APP_SETTINGS = dict(
    # dkron server URL
    URL='http://localhost:8888',
    # used to build browser-visible URLs to dkron - can be a full URL if no reverse proxy is being used
    PATH=None,
    # directory to store and execute the dkron binaries, defaults to temporary one - hardly optimal, do set one up!
    BIN_DIR=None,
    # dkron version to (download and) use
    VERSION='3.2.7',
    # can be changed in case a dkron fork is meant to be used
    DOWNLOAD_URL_TEMPLATE='https://github.com/distribworks/dkron/releases/download/v{version}/dkron_{version}_{system}_{machine}.tar.gz',
    # always `run_dkron` in server mode
    SERVER=False,
    # tags for the agent/server created by `run_dkron` - `label=` tag is not required as it is added by `DKRON_JOB_LABEL`
    TAGS=[],
    # label for the jobs managed by this app, used to make this app agent run only jobs created by this app`
    JOB_LABEL=None,
    # --join when using `run_dkron`
    JOIN=[],
    # workdir of `run_dkron`
    WORKDIR=None,
    # gossip encrypt key for `run_dkron`
    ENCRYPT=None,
    # HTTP Basic auth header value, if dkron instance is protected with it (really recommended, if instance is exposed)
    API_AUTH=None,
    # Token used by `run_dkron` for webhook calls into this app
    TOKEN=None,
    # URL called by dkron webhooks to post job start to this app - passed as `--pre-webhook-url` to dkron, so you need to map `dkron.views.pre_webhook` in your project urls.py and this should be full URL to that route and reachable by dkron. Requires DKRON_SENTRY_CRON_URL otherwise nothing would happen
    PRE_WEBHOOK_URL=None,
    # URL called by dkron webhooks to post job status to this app - passed as `--webhook-url` to dkron, so you need to map `dkron.views.webhook` in your project urls.py and this should be full URL to that route and reachable by dkron
    WEBHOOK_URL=None,
    # string to be prefixed to each job created by this app in dkron so the same dkron cluster can be used by different apps/instances without conflicting job names (assuming unique namespaces ^^)
    NAMESPACE=None,
    # node name to be passed to dkron as `--node-name` - defaults to machine hostname
    NODE_NAME=None,
    # Optional Sentry URL used for monitoring jobs. Use placeholder `<monitor_slug>` in URL for job name.
    SENTRY_CRON_URL=None,
)


class DkronConfig(AppConfig):
    name = 'dkron'
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        for k, v in APP_SETTINGS.items():
            _k = 'DKRON_%s' % k
            if hasattr(settings, _k):
                continue
            if (k, v) == ('PATH', None):
                # special one to default to reverse url
                v = reverse('dkron:proxy')
            setattr(settings, _k, v)



from django.utils import timezone

from logbasecommand.base import LogBaseCommand
from dkron import utils


class Command(LogBaseCommand):
    help = 'Remove old "temp" jobs'

    def add_arguments(self, parser):
        parser.add_argument('-d', '--dry', action='store_true', help='Dry (test) run only, no changes')
        parser.add_argument('-k', '--days', type=int, default=30, help='Number of days to keep')

    def handle(self, *args, **options):
        oldest = int((timezone.now() - timezone.timedelta(days=options['days'])).timestamp())
        self.log(f'Deleting temporary jobs with timestamp lower than {oldest}')
        to_del = []
        jobs = utils._get('jobs', params={'metadata[temp]': 'true'}).json()
        total = len(jobs)
        for j in jobs:
            try:
                ts = int(j['name'].split('_')[-1])
            except Exception:
                # let's see if this happens, it shouldn't...
                self.log_exception('unexpected job name %s - ignoring for now', j['name'])
            if ts < oldest:
                to_del.append(j['name'])

        self.log(f'Deleting {len(to_del)} jobs (out of {total})')
        for x in to_del:
            if options['dry']:
                self.log(f'WOULD delete: {x}')
            else:
                utils.delete_job(x)



from django import forms
from dkron.models import Job


class JobForm(forms.ModelForm):
    def clean_schedule(self):
        data = self.cleaned_data["schedule"]
        if data.startswith("*"):
            raise forms.ValidationError(
                "Job schedule cannot start with * as this will schedule a job to start every second and can have unintended consequences."
            )
        return data

    class Meta:
        model = Job
        fields = "__all__"



# Generated by Django 3.1.5 on 2021-07-29 08:53

from django.db import migrations, models


def data_fix(apps, schema_editor):
    apps.get_model("notifications", "Event").objects.update_or_create(name='dkron_failed_job')


class Migration(migrations.Migration):
    initial = True

    replaces = [
        ('dkron', '0001_initial_20210326'),
        ('dkron', '0004_remove_job_retries'),
        ('dkron', '0001_initial_v2'),
        ('dkron', '0001_initial'),
        ('dkron', '0002_auto_20190103_1758'),
        ('dkron', '0003_auto_20190110_1550'),
        ('dkron', '0004_auto_20190110_1804'),
        ('dkron', '0005_auto_20190213_1700'),
        ('dkron', '0006_job_retries'),
        ('dkron', '0001_squashed_0006_job_retries'),
        ('dkron', '0003_create_notify_event'),
        ('dkron', '0005_auto_20201104_0945'),
        ('dkron', '0002_job_notify_on_error'),
    ]

    dependencies = [
        ('notifications', '0001_initial_20210326'),
    ]

    operations = [
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                (
                    'schedule',
                    models.CharField(
                        help_text='https://dkron.io/usage/cron-spec/ or "@parent JOBNAME" for dependent jobs',
                        max_length=255,
                    ),
                ),
                ('command', models.CharField(max_length=255)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
                ('enabled', models.BooleanField(default=True)),
                ('use_shell', models.BooleanField(default=False, help_text='/bin/sh -c "..."')),
                ('last_run_date', models.DateTimeField(blank=True, editable=False, null=True)),
                ('last_run_success', models.BooleanField(editable=False, null=True)),
                ('notify_on_error', models.BooleanField(default=True)),
            ],
            options={
                'permissions': (('can_use_dashboard', 'Can use the dashboard'),),
            },
        ),
        migrations.RunPython(data_fix),
    ]



from django.urls import path

from . import views

app_name = 'dkron_api'

urlpatterns = [
    path('pre-webhook/', views.pre_webhook, name='pre_webhook'),
    path('webhook/', views.webhook, name='webhook'),
]



from django.urls import path, re_path

from . import views

app_name = 'dkron'

urlpatterns = [
    path('auth/', views.auth, name='auth'),
    path('_/', views.proxy, name='proxy'),
    re_path(r'_/(?P<path>.*)$', views.proxy),
]



from django.db import models


class Job(models.Model):
    name = models.CharField(max_length=255, null=False, blank=False, unique=True)
    schedule = models.CharField(
        max_length=255,
        null=False,
        blank=False,
        help_text='https://dkron.io/docs/usage/cron-spec/ or "@parent JOBNAME" for dependent jobs',
    )
    # if we add more executors (besides the default "shell"), this needs to change
    command = models.CharField(max_length=255, null=False, blank=False)
    description = models.CharField(max_length=255, null=True, blank=True)
    enabled = models.BooleanField(default=True)
    use_shell = models.BooleanField(default=False, help_text='/bin/sh -c "..."')
    last_run_date = models.DateTimeField(null=True, blank=True, editable=False)
    last_run_success = models.BooleanField(null=True, editable=False)
    notify_on_error = models.BooleanField(default=True)
    retries = models.IntegerField(default=0)

    def __str__(self):
        return self.name

    @property
    def namespaced_name(self):
        return utils.add_namespace(self.name)

    @property
    def parent_name(self):
        return self.schedule[8:] if self.schedule.startswith('@parent ') else ''

    class Meta:
        permissions = (("can_use_dashboard", "Can use the dashboard"),)


# circular dependency
from . import utils



from django import template
from django.conf import settings

from dkron import utils

register = template.Library()


@register.simple_tag
def dkron_path():
    return settings.DKRON_PATH


@register.simple_tag
def dkron_executions(job_name):
    return utils.job_executions(job_name)



import tempfile
import os
import requests
import shutil
import tarfile
from pathlib import Path
from django.conf import settings

from logbasecommand.base import LogBaseCommand
from dkron import utils


class Command(LogBaseCommand):
    help = 'Run dkron agent'

    def add_arguments(self, parser):
        parser.add_argument('-s', '--server', action='store_true', help='Run in server mode')
        parser.add_argument('-p', '--http-addr', type=int, default=8888, help='Port used by the web UI')
        parser.add_argument(
            '-j', '--join', action='append', help='Initial agent(s) to join with (can be used multiple times)'
        )
        parser.add_argument(
            '-e', '--encrypt', type=str, help='Key for encrypting network traffic. Must be a base64-encoded 16-byte key'
        )
        parser.add_argument(
            '--node-name',
            type=str,
            default=settings.DKRON_NODE_NAME,
            help='Key for encrypting network traffic. Must be a base64-encoded 16-byte key',
        )

    def download_dkron(self):
        if settings.DKRON_BIN_DIR is None:
            bin_dir = Path(tempfile.mkdtemp())
        else:
            bin_dir = Path(settings.DKRON_BIN_DIR) / settings.DKRON_VERSION

        dl_url, system, _ = utils.dkron_binary_download_url()
        exe_path = bin_dir / ('dkron.exe' if system == 'windows' else 'dkron')

        # check if download is required
        if not exe_path.is_file():
            os.makedirs(bin_dir, exist_ok=True)
            tarball = f'{exe_path}.tar.gz'
            self.log(f'Downloading {dl_url}')
            with requests.get(dl_url, stream=True) as r:
                r.raise_for_status()
                with open(tarball, 'wb') as f:
                    shutil.copyfileobj(r.raw, f)
            tf = tarfile.open(tarball)
            tf.extractall(path=bin_dir)
            os.unlink(tarball)

        return str(exe_path), str(bin_dir)

    def handle(self, *_, **options):
        exe_path, bin_dir = self.download_dkron()

        args = [exe_path, 'agent']

        if options['server'] or settings.DKRON_SERVER:
            args.extend(
                [
                    '--server',
                    '--http-addr',
                    f':{options["http_addr"]}',
                    '--bootstrap-expect',
                    '1',
                    '--data-dir',
                    bin_dir,
                ]
            )
        if options['encrypt'] or settings.DKRON_ENCRYPT:
            args.extend(['--encrypt', options['encrypt'] or settings.DKRON_ENCRYPT])
        if options['node_name']:
            args.extend(['--node-name', options['node_name']])
        for tag in settings.DKRON_TAGS or []:
            args.extend(['--tag', tag])
        # make sure the LABEL tag is included
        if settings.DKRON_JOB_LABEL:
            args.extend(['--tag', f'label={settings.DKRON_JOB_LABEL}'])
        for j in options['join'] or settings.DKRON_JOIN or []:
            args.extend(['--join', j])
        if settings.DKRON_WORKDIR:
            os.chdir(settings.DKRON_WORKDIR)
        if settings.DKRON_PRE_WEBHOOK_URL and settings.DKRON_TOKEN and settings.DKRON_SENTRY_CRON_URL:
            flag_name = '--pre-webhook-url' if utils.dkron_binary_version() < (3, 2, 0) else '--pre-webhook-endpoint'
            args.extend(
                [
                    flag_name,
                    settings.DKRON_PRE_WEBHOOK_URL,
                    '--pre-webhook-payload',
                    f'{settings.DKRON_TOKEN}\n{{{{ .JobName }}}}',
                ]
            )
        if settings.DKRON_WEBHOOK_URL and settings.DKRON_TOKEN:
            flag_name = '--webhook-url' if utils.dkron_binary_version() < (3, 2, 0) else '--webhook-endpoint'
            args.extend(
                [
                    flag_name,
                    settings.DKRON_WEBHOOK_URL,
                    '--webhook-payload',
                    f'{settings.DKRON_TOKEN}\n{{{{ .JobName }}}}\n{{{{ .Success }}}}',
                ]
            )
        os.execv(exe_path, args)



import requests

from django import http
from django.shortcuts import reverse
from django.conf import settings
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from django.contrib.auth.decorators import permission_required

from notifications.utils import notify
from dkron import models, utils


def auth(request):
    # great thing User object is already filled in because Cookies were shared anyway (assuming ProxyPass in nginx)
    # cannot use decorator though because of redirect... fail response needs to be 401 or 403
    # use 401 so nginx redirects to login
    if not request.user.is_authenticated:
        return http.HttpResponse(status=401)
    # and 403 to display access forbidden
    if not request.user.has_perm('dkron.can_use_dashboard'):
        return http.HttpResponseForbidden()
    return http.HttpResponse()


@csrf_exempt
def pre_webhook(request):
    if settings.DKRON_TOKEN is None:
        return http.HttpResponseNotFound()

    if request.method != 'POST':
        return http.HttpResponseBadRequest()

    lines = request.body.decode().splitlines()
    if len(lines) != 2:
        return http.HttpResponseBadRequest()

    if lines[0] != settings.DKRON_TOKEN:
        return http.HttpResponseForbidden()

    job_name = utils.trim_namespace(lines[1])
    if not job_name:
        return http.HttpResponseNotFound()

    o = models.Job.objects.filter(name=job_name).first()
    if o is None:
        return http.HttpResponseNotFound()

    utils.send_sentry_monitor(o, "in_progress")

    return http.HttpResponse()


@csrf_exempt
def webhook(request):
    if settings.DKRON_TOKEN is None:
        return http.HttpResponseNotFound()

    if request.method != 'POST':
        return http.HttpResponseBadRequest()

    lines = request.body.decode().splitlines()
    if len(lines) != 3:
        return http.HttpResponseBadRequest()

    if lines[0] != settings.DKRON_TOKEN:
        return http.HttpResponseForbidden()

    job_name = utils.trim_namespace(lines[1])
    if not job_name:
        return http.HttpResponseNotFound()

    o = models.Job.objects.filter(name=job_name).first()
    if o is None:
        return http.HttpResponseNotFound()

    o.last_run_success = lines[2] == 'true'
    o.last_run_date = timezone.now()
    o.save()

    utils.send_sentry_monitor(o, "ok" if o.last_run_success else "error")

    if not o.last_run_success and o.notify_on_error:
        notify(
            'dkron_failed_job',
            f''':red-pipeline: dkron job *{o.name}* <{request.build_absolute_uri(
                utils.job_executions(o.name)
            )}|failed>''',
        )
    return http.HttpResponse()


@permission_required('dkron.can_use_dashboard')
@csrf_exempt
def proxy(request, path=None):
    """
    reverse proxy implementation based on https://github.com/mjumbewu/django-proxy/blob/master/proxy/views.py
    this is a simplified implementation that "just works" for Dkron but is definitely missing cases for reverse proxying other backends
    re-use in other projects at your own discretion :)

    refer to `dkron authentication` section in docs #FIXME
    """
    headers = {
        k: v
        for k, v in request.headers.items()
        # content-length is not used by requests and might get duplicated (due to casing), just remove it
        # also, dkron does not use cookies, simply drop the whole header to avoid sending django app session over
        if k.lower() not in ('content-length', 'cookie')
    }
    url = utils.dkron_url() + (path or '')

    if settings.DKRON_API_AUTH:
        headers['Authorization'] = f'Basic {settings.DKRON_API_AUTH}'

    response = requests.request(
        request.method,
        url,
        allow_redirects=False,
        headers=headers,
        params=request.GET.copy(),
        data=request.body,
    )

    proxy_response = http.HttpResponse(response.content, status=response.status_code)

    excluded_headers = set(
        [
            # Hop-by-hop headers
            # ------------------
            # Certain response headers should NOT be just tunneled through.  These
            # are they.  For more info, see:
            # http://www.w3.org/Protocols/rfc2616/rfc2616-sec13.html#sec13.5.1
            'connection',
            'keep-alive',
            'proxy-authenticate',
            'proxy-authorization',
            'te',
            'trailers',
            'transfer-encoding',
            'upgrade',
            # Although content-encoding is not listed among the hop-by-hop headers,
            # it can cause trouble as well.  Just let the server set the value as
            # it should be.
            'content-encoding',
            # Since the remote server may or may not have sent the content in the
            # same encoding as Django will, let Django worry about what the length
            # should be.
            'content-length',
        ]
    )
    for key, value in response.headers.items():
        if key.lower() in excluded_headers:
            continue
        elif key.lower() == 'location':
            proxy_response[key] = _fix_location_header(path, value)
        else:
            proxy_response[key] = value

    return proxy_response


def _fix_location_header(path, location):
    base = reverse('dkron:proxy')
    if location.startswith(utils.dkron_url()):
        return base + location[len(utils.dkron_url()) :]
    elif location.startswith('/'):
        return base + location[1:]
    else:
        # this is not meant to cover redirects to "outside" dkron
        # so everything else is caught here
        return base + (path or '') + location



__version__ = '1.2.2'

# set default_app_config when using django earlier than 3.2
try:
    import django

    if django.VERSION < (3, 2):
        default_app_config = 'dkron.apps.DkronConfig'
except ImportError:
    pass
